// Welcome to
// __________         __    __  .__                               __
// \______   \_____ _/  |__/  |_|  |   ____   ______ ____ _____  |  | __ ____
//  |    |  _/\__  \\   __\   __\  | _/ __ \ /  ___//    \\__  \ |  |/ // __ \
//  |    |   \ / __ \|  |  |  | |  |_\  ___/ \___ \|   |  \/ __ \|    <\  ___/
//  |________/(______/__|  |__| |____/\_____>______>___|__(______/__|__\\_____>
//
// This file can be a nice home for your Battlesnake logic and helper functions.
//
// To get you started we've included code to prevent your Battlesnake from moving backwards.
// For more info see docs.battlesnake.com

use std::{cmp::Ordering, time::Instant};

use log::info;
use serde_json::{json, Value};

use crate::{simulation, Action, Battlesnake, Board, Direction, Game, MAX_HEALTH};

impl Battlesnake {
    fn possible_actions_without_heads<'a>(
        &'a self,
        game: &'a Game,
        board: &'a Board,
    ) -> impl Iterator<Item = Direction> + 'a {
        enum_iterator::all::<Direction>()
            .filter(|direction| {
                // filter out directions that would go outside the field
                let target = self.head.move_to(*direction);
                (0..board.width).contains(&target.x) && (0..board.height).contains(&target.y)
            })
            .filter(|direction| {
                let target = self.head.move_to(*direction);
                // don't collide with any snake
                !board
                    .snakes
                    .iter()
                    .filter(|snake| {
                        // filter out snakes that are in our squad if body collisions are allowed
                        !(game.ruleset.settings.squad.allow_body_collisions
                            && self.squad == snake.squad
                            && self.id != snake.id)
                    })
                    .flat_map(|snake| {
                        // get all coordinates of the snake body. The tail of the body can be ignored
                        // if the snake hasn't just eaten, as it will move out of the way
                        let has_eaten = snake.health == MAX_HEALTH;
                        snake.body[..snake.body.len() - usize::from(!has_eaten)].iter()
                    })
                    .any(|&coord| coord == target)
            })
    }

    #[must_use]
    pub fn possible_actions(&self, game: &Game, board: &Board) -> Vec<Direction> {
        self.possible_actions_without_heads(game, board)
            .filter(|direction| {
                // don't go into spots where a bigger snake must go
                let target = self.head.move_to(*direction);
                !board
                    .snakes
                    .iter()
                    .filter(|snake| {
                        // The other snake must be bigger than we are
                        snake.length > self.length
                    })
                    .filter_map(|snake| {
                        // get all snakes movement options
                        let actions = snake
                            .possible_actions_without_heads(game, board)
                            .collect::<Vec<_>>();
                        // only snakes that have a single option
                        match actions[..] {
                            [direction] => Some(snake.head.move_to(direction)),
                            _ => None,
                        }
                    })
                    .any(|coord| coord == target)
            })
            .collect()
    }
}

// info is called when you create your Battlesnake on play.battlesnake.com
// and controls your Battlesnake's appearance
// TIP: If you open your Battlesnake URL in a browser you should see this data
pub fn info() -> Value {
    info!("INFO");

    json!({
        "apiversion": "1",
        "author": "der-informatiker",
        "color": "#00FFEE",
        "head": "smart-caterpillar",
        "tail": "mouse",
    })
}

// start is called when your Battlesnake begins a game
pub fn start(_game: &Game, _turn: i32, _board: &Board, _you: &Battlesnake) {
    info!("GAME START");
}

// end is called when your Battlesnake finishes a game
pub fn end(_game: &Game, _turn: i32, _board: &Board, _you: &Battlesnake) {
    info!("GAME OVER");
}

// move is called on every turn and returns your next move
// Valid moves are "up", "down", "left", or "right"
// See https://docs.battlesnake.com/api/example-move for available data
pub fn get_move(game: &Game, turn: i32, board: &Board, you: &Battlesnake) -> Option<Action> {
    let id_map = board
        .snakes
        .iter()
        .enumerate()
        .map(|(i, snake)| (snake.id.clone(), u8::try_from(i).unwrap()))
        .collect();
    let board = simulation::Board::from_game_board(board, &id_map, turn);

    let my_id = id_map[&you.id];
    let my_index = board.snake_index(my_id)?;

    let possible_actions = board.possible_actions();

    let my_actions = &possible_actions[my_index];

    let actions = my_actions
        .iter()
        .map(|direction| {
            let mut actions = vec![None; possible_actions.len()];
            actions[my_index] = Some(*direction);
            let mut wins = 0;
            let mut total_turns = 0;
            let start = Instant::now();
            for _ in 0..100 {
                let mut board = board.clone();
                board.simulate_with_initial_until(&actions[..], |board| {
                    !board.is_alive(my_id)
                        || (game.ruleset.name != "solo" && board.alive_snakes() <= 1)
                });
                if board.is_alive(my_id) {
                    // we survived
                    wins += 2;
                } else if board.alive_snakes() == 0 {
                    // no snake is alive. This is a draw
                    wins += 1;
                } else {
                    // we lost
                    wins += 0;
                }
                total_turns += board.turn();
            }
            let end = Instant::now();
            info!(
                "Simulation for {direction:?} took {}s",
                (end - start).as_secs_f32()
            );

            (direction, wins, total_turns)
        })
        .collect::<Vec<_>>();

    info!("actions: {actions:?}");

    let (&chosen, _, _) =
        actions
            .into_iter()
            .max_by(
                |(_, score1, turns1), (_, score2, turns2)| match score1.cmp(score2) {
                    Ordering::Equal => turns1.cmp(turns2),
                    order => order,
                },
            )?;

    info!("DIRECTION {}: {:?}", turn, chosen);
    Some(Action {
        r#move: chosen,
        shout: None,
    })
}
